From dddf2f7a1fdc3cff2becf4e4d7adbbf5c0ed31e1 Mon Sep 17 00:00:00 2001
From: Sebastian Pop <spop@amazon.com>
Date: Thu, 5 Aug 2021 23:51:39 +0000
Subject: [PATCH] [arm64] port dequant

dequant_normal		59.29x 	 253.61   	 15036.46
dequant_scaling		44.04x 	 543.91   	 23954.29
quant			24.26x 	 1432.43  	 34753.58
nquant			27.03x 	 1211.85  	 32760.52
---
 source/common/aarch64/asm-primitives.cpp |   5 +
 source/common/aarch64/pixel-util.S       | 134 +++++++++++++++++++++++
 2 files changed, 139 insertions(+)

diff --git a/source/common/aarch64/asm-primitives.cpp b/source/common/aarch64/asm-primitives.cpp
index c1eebf9..2c5ab6c 100644
--- a/source/common/aarch64/asm-primitives.cpp
+++ b/source/common/aarch64/asm-primitives.cpp
@@ -316,6 +316,11 @@ void setupAssemblyPrimitives(EncoderPrimitives &p, int cpuMask)
 
         // quant
         p.quant = PFX(quant_neon);
+        p.nquant = PFX(nquant_neon);
+
+        // dequant_scaling
+        p.dequant_scaling = PFX(dequant_scaling_neon);
+        p.dequant_normal  = PFX(dequant_normal_neon);
 
         p.pu[LUMA_8x4].luma_hvpp   =  interp_8tap_hv_pp_cpu<LUMA_8x4>;
         p.pu[LUMA_8x8].luma_hvpp   =  interp_8tap_hv_pp_cpu<LUMA_8x8>;
diff --git a/source/common/aarch64/pixel-util.S b/source/common/aarch64/pixel-util.S
index 6cf4b26..f0a2ffe 100644
--- a/source/common/aarch64/pixel-util.S
+++ b/source/common/aarch64/pixel-util.S
@@ -1586,6 +1586,96 @@ function x265_pixel_sa8d_64x64_neon
     ret             x10
 endfunc
 
+/***** dequant_scaling*****/
+// void dequant_scaling_c(const int16_t* quantCoef, const int32_t* deQuantCoef, int16_t* coef, int num, int per, int shift)
+function x265_dequant_scaling_neon
+    add             x5, x5, #4              // shift + 4
+    lsr             x3, x3, #3              // num / 8
+    cmp             x5, x4
+    blt             .dequant_skip
+
+    mov             x12, #1
+    sub             x6, x5, x4          // shift - per
+    sub             x6, x6, #1          // shift - per - 1
+    lsl             x6, x12, x6         // 1 << shift - per - 1 (add)
+    dup             v0.4s, w6
+    sub             x7, x4, x5          // per - shift
+    dup             v3.4s, w7
+
+.dequant_loop1:
+    ld1             {v19.8h}, [x0], #16 // quantCoef
+    ld1             {v2.4s}, [x1], #16  // deQuantCoef
+    ld1             {v20.4s}, [x1], #16
+    sub             x3, x3, #1
+    sxtl            v1.4s, v19.4h
+    sxtl2           v19.4s, v19.8h
+
+    mul             v1.4s, v1.4s, v2.4s // quantCoef * deQuantCoef
+    mul             v19.4s, v19.4s, v20.4s
+    add             v1.4s, v1.4s, v0.4s // quantCoef * deQuantCoef + add
+    add             v19.4s, v19.4s, v0.4s
+
+    sshl            v1.4s, v1.4s, v3.4s
+    sshl            v19.4s, v19.4s, v3.4s
+    sqxtn           v16.4h, v1.4s       // x265_clip3
+    sqxtn2          v16.8h, v19.4s
+    st1             {v16.8h}, [x2], #16
+    cbnz            x3, .dequant_loop1
+    ret
+
+.dequant_skip:
+    sub             x6, x4, x5          // per - shift
+    dup             v0.8h, w6
+
+.dequant_loop2:
+    ld1             {v19.8h}, [x0], #16 // quantCoef
+    ld1             {v2.4s}, [x1], #16  // deQuantCoef
+    ld1             {v20.4s}, [x1], #16
+    sub             x3, x3, #1
+    sxtl            v1.4s, v19.4h
+    sxtl2           v19.4s, v19.8h
+
+    mul             v1.4s, v1.4s, v2.4s // quantCoef * deQuantCoef
+    mul             v19.4s, v19.4s, v20.4s
+    sqxtn           v16.4h, v1.4s       // x265_clip3
+    sqxtn2          v16.8h, v19.4s
+
+    sqshl           v16.8h, v16.8h, v0.8h // coefQ << per - shift
+    st1             {v16.8h}, [x2], #16
+    cbnz            x3, .dequant_loop2
+    ret
+endfunc
+
+// void dequant_normal_c(const int16_t* quantCoef, int16_t* coef, int num, int scale, int shift)
+function x265_dequant_normal_neon
+    lsr             w2, w2, #4              // num / 16
+    neg             w4, w4
+    dup             v0.8h, w3
+    dup             v1.4s, w4
+
+.dqn_loop1:
+    ld1             {v2.8h, v3.8h}, [x0], #32
+    smull           v16.4s, v2.4h, v0.4h
+    smull2          v17.4s, v2.8h, v0.8h
+    smull           v18.4s, v3.4h, v0.4h
+    smull2          v19.4s, v3.8h, v0.8h
+
+    srshl           v16.4s, v16.4s, v1.4s
+    srshl           v17.4s, v17.4s, v1.4s
+    srshl           v18.4s, v18.4s, v1.4s
+    srshl           v19.4s, v19.4s, v1.4s
+
+    sqxtn           v2.4h, v16.4s
+    sqxtn2          v2.8h, v17.4s
+    sqxtn           v3.4h, v18.4s
+    sqxtn2          v3.8h, v19.4s
+
+    sub             w2, w2, #1
+    st1             {v2.8h, v3.8h}, [x1], #32
+    cbnz            w2, .dqn_loop1
+    ret
+endfunc
+
 // int psyCost_pp(const pixel* source, intptr_t sstride, const pixel* recon, intptr_t rstride)
 function x265_psyCost_4x4_neon
     ld1r            {v4.2s}, [x0], x1
@@ -1723,3 +1813,47 @@ function x265_quant_neon
     add             w0, w10, w9
     ret
 endfunc
+
+// uint32_t nquant_c(const int16_t* coef, const int32_t* quantCoeff, int16_t* qCoef, int qBits, int add, int numCoeff)
+function x265_nquant_neon
+    neg             x12, x3
+    dup             v0.4s, w12             // q0= -qbits
+    dup             v1.4s, w4              // add
+
+    lsr             w5, w5, #2
+    movi            v4.4s, #0              // v4= accumulate numsig
+    mov             x4, #0
+    movi            v22.4s, #0
+
+.loop_nquant:
+    ld1             {v16.4h}, [x0], #8
+    sub             w5, w5, #1
+    sxtl            v19.4s, v16.4h         // v19 = coef[blockpos]
+
+    cmlt            v18.4s, v19.4s, #0     // v18 = sign
+
+    abs             v19.4s, v19.4s         // v19 = level=abs(coef[blockpos])
+    ld1             {v20.4s}, [x1], #16    // v20 = quantCoeff[blockpos]
+    mul             v19.4s, v19.4s, v20.4s // v19 = tmplevel = abs(level) * quantCoeff[blockpos];
+
+    add             v20.4s, v19.4s, v1.4s  // v20 = tmplevel+add
+    sshl            v20.4s, v20.4s, v0.4s  // v20 = level =(tmplevel+add) >> qbits
+
+    // numsig
+    cmeq            v21.4s, v20.4s, v22.4s
+    add             v4.4s, v4.4s, v21.4s
+    add             x4, x4, #4
+
+    eor             v21.16b, v20.16b, v18.16b
+    sub             v21.4s, v21.4s, v18.4s
+    sqxtn           v16.4h, v21.4s
+    abs             v17.4h, v16.4h
+    st1             {v17.4h}, [x2], #8
+
+    cbnz            w5, .loop_nquant
+
+    uaddlv          d4, v4.4s
+    fmov            x12, d4
+    add             x0, x4, x12
+    ret
+endfunc
-- 
2.25.1

